#include "displayspace.h"

#include <iostream>
#include <fstream>

#include <boost/filesystem.hpp>
using namespace boost;

using namespace cv;
using namespace std;
using namespace gaze_data;

DisplaySpace::DisplaySpace(string calib_file, string display_file)
{
    // read calibration information
    if(!filesystem::exists(filesystem::path(calib_file)))
    {
        cout << "Calibration data not found" << endl;
        exit(EXIT_FAILURE);
    }

    FileStorage fs_calib(calib_file, FileStorage::READ);

    fs_calib["camera_matrix"] >> camera_matrix_;
    fs_calib["dist_coeffs"] >> dist_coeffs_;

    fs_calib.release();

    // read monitor information
    if(!filesystem::exists(filesystem::path(display_file)))
    {
        cout << "Display pose data not found" << endl;
        exit(EXIT_FAILURE);
    }

    FileStorage fs_disp(display_file, FileStorage::READ);

    fs_disp["monitor_W"] >> monitor_W_;
    fs_disp["monitor_H"] >> monitor_H_;
    fs_disp["monitor_R"] >> monitor_R_;
    fs_disp["monitor_T"] >> monitor_T_;

    // compute monitor plane
    Vec3f corners[4];
    corners[0] = Vec3f(0.0, 0.0, 0.0);
    corners[1] = Vec3f(monitor_W_, 0.0, 0.0);
    corners[2] = Vec3f(0.0, monitor_H_, 0.0);
    corners[3] = Vec3f(monitor_W_, monitor_H_, 0.0);

    for(int i=0; i<4; i++)
    {
        Mat corners_cam = monitor_R_ * Mat(corners[i]) + monitor_T_;
        corners_cam.copyTo(monitor_corners_[i]);
    }

    Vec3f normal = Vec3f(0.0, 0.0, 1.0); // normal direction
    Mat normal_cam = monitor_R_ * Mat(normal);
    normal_cam.copyTo(monitor_normal_);
}

DisplaySpace::~DisplaySpace()
{
}

void DisplaySpace::project_to_display(vector<eyeinfo> &eyes, vector<gazeinfo> &gazes, Mat &debug_img) // for the eye model
{
    for(size_t e=0; e<eyes.size(); ++e)
    {
        gazeinfo gaze_t;
        eyeinfo eye_t = eyes[e];

        gaze_t.face_id = eye_t.face_id;

        intersect_gaze(eye_t.leye_pos, eye_t.lgaze_vec, gaze_t.lgaze);
        intersect_gaze(eye_t.reye_pos, eye_t.rgaze_vec, gaze_t.rgaze);

        gazes.push_back(gaze_t);
    }

    if(!debug_img.empty())
    {
        static const int dW = 640;
        static const int dH = 360;
        Mat debug_disp = Mat::zeros(Size(dW, dH), CV_8UC3);

        for(size_t g=0; g<gazes.size(); ++g)
        {
            Point2f gl = gazes[g].lgaze;
            gl.x = dW*gl.x/monitor_W_;
            gl.y = dH*gl.y/monitor_H_;

            Point2f gr = gazes[g].rgaze;
            gr.x = dW*gr.x/monitor_W_;
            gr.y = dH*gr.y/monitor_H_;

            circle(debug_disp, gl, 2, CV_RGB(0,0,255), -1);
            circle(debug_disp, gr, 2, CV_RGB(255,0,0), -1);
            circle(debug_disp, 0.5*(gl+gr), 5, CV_RGB(255,0,255), -1);
        }

        debug_disp.copyTo(debug_img(Rect(0, 0, dW, dH)));
    }
}

void DisplaySpace::project_to_display(vector<facepatchinfo> &facepatches, vector<gazeinfo> &gazes, Mat &debug_img) // for the face model
{
    for(size_t f=0; f<facepatches.size(); ++f)
    {
        gazeinfo gaze_t;
        facepatchinfo face_patch = facepatches[f];

        gaze_t.face_id = face_patch.face_id;

        intersect_gaze(face_patch.face_pos, face_patch.gaze_vec, gaze_t.gaze);

        gazes.push_back(gaze_t);
    }

    if(!debug_img.empty())
    {
        static const int dW = 640;
        static const int dH = 360;
        Mat debug_disp = Mat::zeros(Size(dW, dH), CV_8UC3);

        for(size_t g=0; g<gazes.size(); ++g)
        {
            Point2f gl = gazes[g].lgaze;
            gl.x = dW*gl.x/monitor_W_;
            gl.y = dH*gl.y/monitor_H_;

            Point2f gr = gazes[g].rgaze;
            gr.x = dW*gr.x/monitor_W_;
            gr.y = dH*gr.y/monitor_H_;

            circle(debug_disp, gl, 2, CV_RGB(0,0,255), -1);
            circle(debug_disp, gr, 2, CV_RGB(255,0,0), -1);
            circle(debug_disp, 0.5*(gl+gr), 5, CV_RGB(255,0,255), -1);
        }

        debug_disp.copyTo(debug_img(Rect(0, 0, dW, dH)));
    }
}

void DisplaySpace::intersect_gaze(const Vec3f origin, const Vec3f gaze_vec, Point2f &gaze_pos)
{
    // compute intersection
    float gaze_len = Mat(monitor_normal_).dot(Mat(monitor_corners_[0]-origin))/Mat(monitor_normal_).dot(Mat(gaze_vec));
    Vec3f gaze_pos_cam = origin + gaze_len * gaze_vec;

    // convert to monitor coodinate system
    Mat gaze_pos_ = monitor_R_.t() * (Mat(gaze_pos_cam) - monitor_T_);
    Vec3f gaze_pos_3d;
    gaze_pos_.copyTo(gaze_pos_3d);

    gaze_pos.x = gaze_pos_3d.val[0];
    gaze_pos.y = gaze_pos_3d.val[1];
}

void DisplaySpace::debug_show(Mat &debug_img, vector<eyeinfo> &eyes)
{
    static const float l_dist = 500.0;

    Mat zero = Mat::zeros(1, 3, CV_32F);

    for(size_t e=0; e<eyes.size(); ++e)
    {
        eyeinfo eye_t = eyes[e];
        Vec3f eo = eye_t.leye_pos;
        Vec3f gv = eye_t.lgaze_vec;

        Vec3f v0 = monitor_corners_[0]-eo;
        Vec3f v1 = monitor_corners_[1]-eo;
        Vec3f v2 = monitor_corners_[2]-eo;
        Vec3f v3 = monitor_corners_[3]-eo;

        v0 = eo + l_dist*v0/norm(v0);
        v1 = eo + l_dist*v1/norm(v1);
        v2 = eo + l_dist*v2/norm(v2);
        v3 = eo + l_dist*v3/norm(v3);

        Vec3f nv = (v1-v0).cross(v2-v0);
        nv = nv/norm(nv);

        float gl = nv.dot(v0 - eo)/nv.dot(gv);
        Vec3f gp = eo + gl * gv;

        vector<Point3f> cam_points;
        cam_points.push_back(eo);
        cam_points.push_back(v0);
        cam_points.push_back(v1);
        cam_points.push_back(v2);
        cam_points.push_back(v3);
        cam_points.push_back(gp);;
        vector<Point2f> img_points;
        projectPoints(cam_points, zero, zero, camera_matrix_, dist_coeffs_, img_points);
    }
}
