#include "facetracker.h"

#include <tbb/tbb.h>

#include <boost/filesystem.hpp>
using namespace boost::filesystem;

using namespace cv;
using namespace std;
using namespace gaze_data;

// (currently) hard-coded setting
const int FaceTracker::num_faces_max_ = 8;
const float FaceTracker::detection_resize_rate_ = 2.0;
const int FaceTracker::detection_skip_frames_ = 1;
const float FaceTracker::nonoverlap_threshold_ = 0.5;
const double FaceTracker::certainty_threshold_ = -0.1; // the smaller the better
const int FaceTracker::landmark_indices_[6] = {36, 39, 42, 45, 48, 54};
const int FaceTracker::tracking_loss_limit_ = 5;

FaceTracker::FaceTracker(string model_dir)
{
    // initialize the tracking models
    LandmarkDetector::FaceModelParameters det_parameter_;

    path root = model_dir;
    det_parameter_.model_location = (root / "model/main_clm_wild.txt").string();
    det_parameter_.face_detector_location = (root / "classifiers/haarcascade_frontalface_alt.xml").string();// this line will be disable due to "curr_face_detector"

    det_parameter_.use_face_template = true;
    det_parameter_.reinit_video_every = 5;
    det_parameter_.quiet_mode = true;
    // For in the wild fitting these parameters are suitable
    det_parameter_.window_sizes_init = vector<int>(4);
    det_parameter_.window_sizes_init[0] = 15; det_parameter_.window_sizes_init[1] = 13; det_parameter_.window_sizes_init[2] = 11; det_parameter_.window_sizes_init[3] = 9;
    det_parameter_.sigma = 1.25;
    det_parameter_.reg_factor = 35;
    det_parameter_.weight_factor = 2.5;
    det_parameter_.num_optimisation_iteration = 10;   
    det_parameter_.curr_face_detector = LandmarkDetector::FaceModelParameters::HOG_SVM_DETECTOR;
    

    det_parameters_.push_back(det_parameter_);

    LandmarkDetector::CLNF clnf_model_ = LandmarkDetector::CLNF(det_parameter_.model_location);
    clnf_model_.face_detector_HAAR.load(det_parameter_.face_detector_location);
    clnf_model_.face_detector_location = det_parameter_.face_detector_location;
    clnf_models_.reserve(num_faces_max_);
    clnf_models_.push_back(clnf_model_);
    active_models_.push_back(false);

    for(int i=1; i<num_faces_max_; ++i)
    {
        clnf_models_.push_back(clnf_model_);
        active_models_.push_back(false);
        det_parameters_.push_back(det_parameter_);
    }

    // variables
    frame_counter_ = 0;
    current_face_id_ = 1;
    for(int i=0; i<num_faces_max_; ++i) face_ids_.push_back(0);
}

FaceTracker::~FaceTracker()
{
}

void FaceTracker::reset()
{
    // reset all status
    frame_counter_ = 0;
    current_face_id_ = 1;

    for(unsigned int model = 0; model < clnf_models_.size(); ++model)
    {
        active_models_[model] = false;
        face_ids_[model] = 0;
        clnf_models_[model].Reset();
    }
}

void FaceTracker::track_faces(Mat &input_img, vector<faceinfo> &results, Mat &debug_img)
{
    if(input_img.channels() != 3)
    {
        cout << "The input must be a color image!" <<endl;
        getchar();
        exit(EXIT_FAILURE);
    }

    Mat_<float> depth_image;
    Mat grayscale_image;
    cvtColor(input_img, grayscale_image, CV_BGR2GRAY);

    bool all_models_active = true;
    for(unsigned int model = 0; model < clnf_models_.size(); ++model)
    {
        if(!active_models_[model])
        {
            all_models_active = false;
            break;
        }
    }

    // Detect faces
    // Get the detections (every Xth frame and when there are free models available for tracking)
    std::vector<Rect_<double> > face_detections_;
    if(frame_counter_ % detection_skip_frames_ == 0 && !all_models_active)
    {
        // resized image for faster face detection
        cv::Mat small_grayscale_image_;
        resize(grayscale_image, small_grayscale_image_,
               Size(), 1.0/detection_resize_rate_, 1.0/detection_resize_rate_);

        if(det_parameters_[0].curr_face_detector == LandmarkDetector::FaceModelParameters::HOG_SVM_DETECTOR)
        {
            vector<double> confidences;
            LandmarkDetector::DetectFacesHOG(face_detections_, small_grayscale_image_, clnf_models_[0].face_detector_HOG, confidences);
        }
        else
        {
            LandmarkDetector::DetectFaces(face_detections_, small_grayscale_image_, clnf_models_[0].face_detector_HAAR);
        }

        // resize the face deteciton back
        for(int num_f=0; num_f<(int)face_detections_.size(); ++num_f)
        {
            face_detections_[num_f].x = face_detections_[num_f].x * detection_resize_rate_;
            face_detections_[num_f].y = face_detections_[num_f].y * detection_resize_rate_;
            face_detections_[num_f].width = face_detections_[num_f].width * detection_resize_rate_;
            face_detections_[num_f].height = face_detections_[num_f].height * detection_resize_rate_;
        }

        if(!debug_img.empty())
        {
            for(int num_f=0; num_f<(int)face_detections_.size(); ++num_f)
                rectangle(debug_img, face_detections_[num_f], CV_RGB(255,0,0));
        }

        // Keep only non overlapping detections (also convert to a concurrent vector
        NonOverlapingDetections(clnf_models_, face_detections_);
    }

    vector< tbb::atomic<bool> > face_detections_used(face_detections_.size());

    // Go through every model and update the tracking
    tbb::parallel_for(0, (int)clnf_models_.size(), [&](int model)
    {
        // If the current model has failed more than threshold, remove it
        if(clnf_models_[model].failures_in_a_row > tracking_loss_limit_)
        {
            active_models_[model] = false;
            face_ids_[model] = 0;
            clnf_models_[model].Reset();
        }

        // If the model is inactive reactivate it with new detections
        if(!active_models_[model])
        {
            for(size_t detection_ind = 0; detection_ind < face_detections_.size(); ++detection_ind)
            {
                // if it was not taken by another tracker take it (if it is false swap it to true and enter detection, this makes it parallel safe)
                if(face_detections_used[detection_ind].compare_and_swap(true, false) == false)
                {
                    // Reinitialise the model
                    clnf_models_[model].Reset();
                    // This ensures that a wider window is used for the initial landmark localisation
                    clnf_models_[model].detection_success = false;
                    LandmarkDetector::DetectLandmarksInVideo(grayscale_image, depth_image, face_detections_[detection_ind], clnf_models_[model], det_parameters_[model]);
                    // This activates the model
                    active_models_[model] = true;
                    face_ids_[model] = current_face_id_;
                    current_face_id_++;
                    // break out of the loop as the tracker has been reinitialised
                    break;
                }
            }
        }
        else
        {
            // The actual facial landmark detection / tracking
            LandmarkDetector::DetectLandmarksInVideo(grayscale_image, depth_image, clnf_models_[model], det_parameters_[model]);
        }
    });

    // Go through every model and check the results
    for(size_t model=0; model<clnf_models_.size(); ++model)
    {
        // Check if the alignment result is overlapping previous models
        bool overlapping = false;
        for(size_t model_ref=0; model_ref<model; ++model_ref)
        {
            double overlap_ratio = NonOverlapingDetection(clnf_models_[model_ref], clnf_models_[model]);
            if(overlap_ratio > nonoverlap_threshold_) overlapping = true;
        }

        if(overlapping)
        {
            active_models_[model] = false;
            face_ids_[model] = 0;
            clnf_models_[model].Reset();
            continue;
        }

        double detection_certainty = clnf_models_[model].detection_certainty;

        if(detection_certainty > certainty_threshold_) continue;

        faceinfo face;
        face.certainty = clnf_models_[model].detection_certainty;
        face.face_id = face_ids_[model];
        face.face_bb = clnf_models_[model].GetBoundingBox();

        for(int p=0; p<6; p++)
        {
            int num_p = landmark_indices_[p];
            face.landmarks[p] = Point2d(
                        clnf_models_[model].detected_landmarks.at<double>(num_p,0),
                        clnf_models_[model].detected_landmarks.at<double>(num_p+68,0)
                        );
        }

        if(!debug_img.empty())
        {
            for(int p=0; p<6; p++)
                circle(debug_img, Point(face.landmarks[p].x, face.landmarks[p].y), 3, CV_RGB(0,255,0), -1);
            putText(debug_img, std::to_string(face.face_id),
                    Point(face.landmarks[3].x + 10, face.landmarks[3].y),
                    FONT_HERSHEY_DUPLEX, 1.0, CV_RGB(255,255,255));
        }

        results.push_back(face);
    }

    frame_counter_++;
}

// merge the face deteciton result
void FaceTracker::NonOverlapingDetections(const vector<LandmarkDetector::CLNF>& CLNF_models, vector<Rect_<double> >& face_detections)
{
    // Go over the model and eliminate detections that are not informative (there already is a tracker there)
    for(size_t model = 0; model < CLNF_models.size(); ++model){
        // See if the detections intersect
        Rect_<double> model_rect = CLNF_models[model].GetBoundingBox();

        for(int detection = face_detections.size()-1; detection >=0; --detection){
            double intersection_area = (model_rect & face_detections[detection]).area();
            double union_area = model_rect.area() + face_detections[detection].area() - 2 * intersection_area;

            // If the model is already tracking what we're detecting ignore the detection, this is determined by amount of overlap
            if( intersection_area/union_area > nonoverlap_threshold_)
            {
                face_detections.erase(face_detections.begin() + detection);
            }
        }
    }
}

double FaceTracker::NonOverlapingDetection(const LandmarkDetector::CLNF &ref_model, const LandmarkDetector::CLNF &tgt_model)
{
    Rect_<double> ref_rect = ref_model.GetBoundingBox();
    Rect_<double> tgt_rect = tgt_model.GetBoundingBox();

    double intersection_area = (ref_rect & tgt_rect).area();
    double union_area = ref_rect.area() + tgt_rect.area() - 2 * intersection_area;

    return intersection_area/union_area;
}
