#include "gazeestimator.h"
#include <cmath>

using namespace std;
using namespace cv;
using namespace caffe;
using namespace gaze_data;

GazeEstimator::GazeEstimator(string param_path, string model_path)
{
    p_net_ = NULL;

    // Set GPU (or CPU)
    caffe::Caffe::set_mode(caffe::Caffe::GPU);

    // create CNN
    p_net_ = new Net<float>(param_path, caffe::TEST);

    // load pre-trained weights (binary proto)
    p_net_->CopyTrainedLayersFrom(model_path);

    // judge model type base on the paramater file name
    size_t i = param_path.rfind("/", param_path.length());
    string filename;
    if (i != string::npos) {
       filename = param_path.substr(i+1, param_path.length() - i);
    }
    if (!filename.compare(string("lenet_test.prototxt")))
    {model_type_ = 1;}
    else if (!filename.compare(string("googlenet.prototxt")))
    {model_type_ = 2;}
    else if (!filename.compare(string("alexnet_eye.prototxt")))
    {model_type_ = 3;}
    else if (!filename.compare(string("alexnet_face_448.prototxt")))
    {model_type_ = 4;} // the single face model
    else if (!filename.compare(string("alexnet_face.prototxt")))
    {model_type_ = 4;} // the single face model
    else
    {
        model_type_ = 0;
        cout<<"Cannot define the type of model!"<<endl;
        exit(EXIT_FAILURE);
    }
}

GazeEstimator::~GazeEstimator()
{
    if(p_net_ != NULL) delete p_net_;
}

void GazeEstimator::estimate(vector<eyeinfo> &eyes)
{
    // TODO: parallelize
    for(size_t e=0; e<eyes.size(); ++e)
    {
        // preprocess features
        Mat leye, reye;
        Vec2f lrot, rrot;
        prepare_data(eyes[e].leye_img, eyes[e].leye_rot, leye, lrot, false);
        prepare_data(eyes[e].reye_img, eyes[e].reye_rot, reye, rrot, true);

        // estimation using CNN
        process_data(leye, lrot, eyes[e].lgaze_angle, 60, 36);
        process_data(reye, rrot, eyes[e].rgaze_angle, 60, 36 );

        // convert to unit gaze vector (already in the camera coordinate system)
        convert_gaze(eyes[e].lgaze_angle, eyes[e].leye_conv, eyes[e].lgaze_vec, false);
        convert_gaze(eyes[e].rgaze_angle, eyes[e].reye_conv, eyes[e].rgaze_vec, true);
    }
}

void GazeEstimator::estimate(vector<facepatchinfo> &facepatches)
{
    // TODO: parallelize
    for(size_t f=0; f<facepatches.size(); ++f)
    {
        // preprocess features
        Mat face;
        Vec2f rot;
        prepare_data(facepatches[f].face_img, facepatches[f].face_rot, face, rot);
        // estimation using CNN
        process_data(face, rot, facepatches[f].gaze_angle, 448, 448, facepatches[f].feature); //

        // convert to unit gaze vector (in the camera coordinate system)
        convert_gaze(facepatches[f].gaze_angle, facepatches[f].face_conv, facepatches[f].gaze_vec, false);
    }
}

void GazeEstimator::prepare_data(const Mat &src_eye, const Mat &src_rot,
                                 Mat &dst_eye, Vec2f &dst_rot, const bool is_right)
{
    // copy eye image
    Mat eye_pose = src_rot.clone();
    if(is_right)
    {
        // right->left flip
        flip(src_eye, dst_eye, 1);
        eye_pose.at<float>(1,0) = -1.0 * eye_pose.at<float>(1,0);
    }
    else dst_eye = src_eye.clone();

    // angle representation
    Mat M;
    Rodrigues(eye_pose, M);
    Mat Zv = M.col(2);
    float theta = asin(Zv.at<float>(1,0));
    float phi = atan2(Zv.at<float>(0,0), Zv.at<float>(2,0));
    dst_rot = Vec2f(theta, phi);
}

void GazeEstimator::prepare_data(const Mat &src_face, const Mat &src_rot,
                                 Mat &dst_face, Vec2f &dst_rot)
{
    // copy eye image
    Mat face_pose = src_rot.clone();
    dst_face = src_face.clone();

    // angle representation
    Mat M;
    Rodrigues(face_pose, M);
    Mat Zv = M.col(2);
    float theta = asin(Zv.at<float>(1,0));
    float phi = atan2(Zv.at<float>(0,0), Zv.at<float>(2,0));
    dst_rot = Vec2f(theta, phi);
}

void GazeEstimator::process_data(const Mat &img, const Vec2f &rot, Vec2f &gaze, int img_width, int img_height, float* feature)
{
    vector<Mat> img_vec;
    cv::Mat img_resize;
    resize(img, img_resize, Size(img_width, img_height));

    img_vec.push_back(img_resize);

    std::vector<int> labelVector;
    labelVector.clear();
    labelVector.push_back(1);
    labelVector.push_back(1);
    float loss = 0.0;
    caffe::shared_ptr<caffe::MemoryDataLayer<float> > data_layer_;
    data_layer_ = boost::static_pointer_cast<MemoryDataLayer<float> >(p_net_->layer_by_name("data"));
    data_layer_->AddMatVector(img_vec, labelVector);

    // run network
    p_net_->ForwardPrefilled(&loss);

    if (model_type_==1)
    {
        // get output layer "ip2"
        float *temp = (float*)p_net_->blob_by_name("ip2")->cpu_data();
        // copy estimated gaze vector
        gaze.val[0] = temp[0];
        gaze.val[1] = temp[1];
        temp = NULL;
    }
    else if (model_type_==2)// if it is googlenet
    {
        float *temp1 = (float*)p_net_->blob_by_name("loss1/classifier")->cpu_data();
        float *temp2 = (float*)p_net_->blob_by_name("loss2/classifier")->cpu_data();
        float *temp3 = (float*)p_net_->blob_by_name("loss3/classifier")->cpu_data();
        // average the output of three output values
        gaze.val[0] = (temp1[0]+temp2[0]+temp3[0]) / 3.0f;
        gaze.val[1] = (temp1[1]+temp2[1]+temp3[1]) / 3.0f;
        temp1 = NULL;
        temp2 = NULL;
        temp3 = NULL;
    }
    else if (model_type_==3)// if it is alexnet
    {
        float *temp = (float*)p_net_->blob_by_name("fc8")->cpu_data();
        // copy estimated gaze vector
        gaze.val[0] = temp[0];
        gaze.val[1] = temp[1];
    }
    else if (model_type_==4)// if it is alexnet
    {
        float *temp = (float*)p_net_->blob_by_name("fc8")->cpu_data();
        // copy estimated gaze vector
        gaze.val[0] = temp[0];
        gaze.val[1] = temp[1];
        temp = (float*)p_net_->blob_by_name("fc6_gaze")->cpu_data();
        for (int num_f=0; num_f<4096; ++num_f)
        {
            feature[num_f] = temp[num_f];
        }
        temp = NULL;
    }
}

void GazeEstimator::convert_gaze(const Vec2f &gaze_angle, const Mat &cnv_mat,
                                 Vec3f &gaze_vec, const bool is_right)
{
    float theta = gaze_angle.val[0];
    float phi = gaze_angle.val[1];

    float gaze_x = (-1.0f)*cos(theta)*sin(phi);
    float gaze_y = (-1.0f)*sin(theta);
    float gaze_z = (-1.0f)*cos(theta)*cos(phi);

    // flip the gaze vector here too
    if(is_right) gaze_x *= -1.0;

    // convert to the original camera coordinate system
    Vec3f gaze_v(gaze_x, gaze_y, gaze_z);

    // apply denormalization
    Mat gaze_v_cam = cnv_mat.inv() * Mat(gaze_v);
    gaze_v_cam = gaze_v_cam / norm(gaze_v_cam);
    // copy
    gaze_v_cam.copyTo(gaze_vec);
}
