#include "headposeestimator.h"

#include <iostream>
#include <fstream>

#include <boost/filesystem.hpp>
using namespace boost;

using namespace cv;
using namespace std;
using namespace gaze_data;

HeadPoseEstimator::HeadPoseEstimator(string calib_file, string model_file)
{
    // read calibration information
    if(!filesystem::exists(filesystem::path(calib_file)))
    {
        cout << "Calibration data not found" << endl;
        exit(EXIT_FAILURE);
    }

    FileStorage fs_calib(calib_file, FileStorage::READ);

    fs_calib["camera_matrix"] >> camera_matrix_;
    fs_calib["dist_coeffs"] >> dist_coeffs_;

    fs_calib.release();

    // read face model
    if(!filesystem::exists(filesystem::path(model_file)))
    {
        cout << "Face model data not found" << endl;
        exit(EXIT_FAILURE);
    }

    FileStorage fs_model(model_file, FileStorage::READ);

    fs_model["face_model"] >> face_model_mat_;

    for(int p=0; p<6; ++p)
        face_model_.push_back(Point3f(face_model_mat_.at<float>(0,p),
                                      face_model_mat_.at<float>(1,p),
                                      face_model_mat_.at<float>(2,p)));

    fs_model.release();
}

HeadPoseEstimator::~HeadPoseEstimator()
{
}

void HeadPoseEstimator::obtain_eyepatches(Mat &input_img, vector<faceinfo> &landmarks,
                                          vector<eyeinfo> &results, Mat &debug_img)
{
    Mat undist_img;
    undistort(input_img, undist_img, camera_matrix_, dist_coeffs_);

    // TODO: parallelize
    for(size_t face=0; face<landmarks.size(); ++face)
    {
        eyeinfo eye_t;
        faceinfo face_t = landmarks[face];
        eye_t.face_id = face_t.face_id;

        vector<Point2f> landmarks_orig(face_t.landmarks,
                                       face_t.landmarks + sizeof(face_t.landmarks)/ sizeof(face_t.landmarks[0]));

        vector<Point2f> landmarks_undist;
        undistortPoints(landmarks_orig, landmarks_undist, camera_matrix_, dist_coeffs_, Mat(), camera_matrix_);

        // estimate head pose via model fitting
        Mat zero_dist = Mat::zeros(1, 5, CV_32F);
        Mat rvec, tvec;
        solvePnP(face_model_, landmarks_undist, camera_matrix_, zero_dist, rvec, tvec, false, SOLVEPNP_DLS);
        solvePnP(face_model_, landmarks_undist, camera_matrix_, zero_dist, rvec, tvec, true);
        rvec.convertTo(eye_t.head_r, CV_32F);
        tvec.convertTo(eye_t.head_t, CV_32F);

        // obtain normalized eye images and corresponding pose features
        get_normalized_data(undist_img, eye_t);

        if(!debug_img.empty())
        {
            draw_landmarks(debug_img, rvec, tvec);
            draw_headpose(debug_img, rvec, tvec);

            Mat leye, reye;
            cvtColor(eye_t.leye_img, leye, CV_GRAY2BGR);
            cvtColor(eye_t.reye_img, reye, CV_GRAY2BGR);
            Rect leye_roi(landmarks_undist[3].x+10+leye.cols, landmarks_undist[3].y-leye.rows,
                          leye.cols, leye.rows);
            if (leye_roi.x>0 && (leye_roi.x+leye_roi.width) < debug_img.cols && leye_roi.y>0 && (leye_roi.y+leye_roi.height)<debug_img.rows )
                leye.copyTo(debug_img(leye_roi));

            Rect reye_roi(landmarks_undist[3].x+10, landmarks_undist[3].y-leye.rows,
                      leye.cols, leye.rows);
            if (reye_roi.x>0 && (reye_roi.x+reye_roi.width) < debug_img.cols && reye_roi.y>0 && (reye_roi.y+reye_roi.height)<debug_img.rows )
                reye.copyTo(debug_img(reye_roi));
        }

        results.push_back(eye_t);
    }
}

void HeadPoseEstimator::obtain_facepatches(Mat &input_img, vector<faceinfo> &landmarks,
                                          vector<facepatchinfo> &results, Mat &debug_img)
{
    Mat undist_img;
    undistort(input_img, undist_img, camera_matrix_, dist_coeffs_);

    // TODO: parallelize
    for(size_t f=0; f<landmarks.size(); ++f)
    {
        facepatchinfo facepatch;
        faceinfo face_t = landmarks[f];
        facepatch.face_id = face_t.face_id;
        vector<Point2f> landmarks_orig(face_t.landmarks,
                                       face_t.landmarks + sizeof(face_t.landmarks)/ sizeof(face_t.landmarks[0]));
        vector<Point2f> landmarks_undist;
        undistortPoints(landmarks_orig, landmarks_undist, camera_matrix_, dist_coeffs_, Mat(), camera_matrix_);

        // estimate head pose via model fitting
        Mat zero_dist = Mat::zeros(1, 5, CV_32F);
        Mat rvec, tvec;
        solvePnP(face_model_, landmarks_undist, camera_matrix_, zero_dist, rvec, tvec, false, SOLVEPNP_DLS);
        solvePnP(face_model_, landmarks_undist, camera_matrix_, zero_dist, rvec, tvec, true);
        rvec.convertTo(facepatch.head_r, CV_32F);
        tvec.convertTo(facepatch.head_t, CV_32F);

        // obtain normalized eye images and corresponding pose features
        get_normalized_face(undist_img, facepatch);

        if(!debug_img.empty())
        {
            draw_landmarks(debug_img, rvec, tvec);
            draw_headpose(debug_img, rvec, tvec);

            Rect face_roi(landmarks_undist[1].x+50, landmarks_undist[3].y, facepatch.face_img.cols, facepatch.face_img.rows);
            if (face_roi.x>0 && (face_roi.x+face_roi.width) < debug_img.cols && face_roi.y>0 && (face_roi.y+face_roi.height)<debug_img.rows )
                facepatch.face_img.copyTo(debug_img(face_roi));
        }

        results.push_back(facepatch);
    }
}

void HeadPoseEstimator::draw_landmarks(Mat &img, const Mat &rvec, const Mat &tvec)
{
    vector<Point2f> landmarks;
    projectPoints(face_model_, rvec, tvec, camera_matrix_, dist_coeffs_, landmarks);

    for(int p=0; p<6; p++) circle(img, landmarks[p], 5, CV_RGB(0,255,0), -1);
}

void HeadPoseEstimator::draw_headpose(Mat &img, const Mat &rvec, const Mat &tvec)
{
    vector<Point3f> model_axis;
    model_axis.push_back(Point3f( 0.0f,-20.0f,  0.0f));
    model_axis.push_back(Point3f(50.0f,-20.0f,  0.0f));
    model_axis.push_back(Point3f( 0.0f,-70.0f,  0.0f));
    model_axis.push_back(Point3f( 0.0f,-20.0f,-50.0f));

    vector<Point2f> img_axis;
    projectPoints(model_axis, rvec, tvec, camera_matrix_, dist_coeffs_, img_axis);

    line(img, img_axis[0], img_axis[1], CV_RGB(0,255,255), 2);
    line(img, img_axis[0], img_axis[2], CV_RGB(0,255,255), 2);
    line(img, img_axis[0], img_axis[3], CV_RGB(0,255,255), 2);
}

void HeadPoseEstimator::draw_gaze(Mat &debug_img, vector<eyeinfo> &eyes)
{
    static const float gaze_length = 300.0;

    Mat zero = Mat::zeros(1, 3, CV_32F);

    for(size_t e=0; e<eyes.size(); ++e)
    {
        eyeinfo eye_t = eyes[e];

        Mat rvec, tvec;
        eye_t.head_r.convertTo(rvec, CV_32F);
        eye_t.head_t.convertTo(tvec, CV_32F);
        draw_landmarks(debug_img, rvec, tvec);

        vector<Point3f> cam_points;
        cam_points.push_back(eye_t.leye_pos);
        cam_points.push_back(eye_t.leye_pos + gaze_length*eye_t.lgaze_vec);
        cam_points.push_back(eye_t.reye_pos);
        cam_points.push_back(eye_t.reye_pos + gaze_length*eye_t.rgaze_vec);

        vector<Point2f> img_points;
        projectPoints(cam_points, zero, zero, camera_matrix_, dist_coeffs_, img_points);

        line(debug_img, img_points[0], img_points[1], CV_RGB(255,0,0), 5);
        line(debug_img, img_points[2], img_points[3], CV_RGB(255,0,0), 5);
        circle(debug_img, img_points[1], 3, CV_RGB(255,0,0), -1);
        circle(debug_img, img_points[3], 3, CV_RGB(255,0,0), -1);
    }
}

void HeadPoseEstimator::draw_gaze(Mat &debug_img, vector<facepatchinfo> &facepathces)
{
    static const float gaze_length = 300.0;

    Mat zero = Mat::zeros(1, 3, CV_32F);

    for(size_t f=0; f<facepathces.size(); ++f)
    {
        facepatchinfo face_patch = facepathces[f];

        Mat rvec, tvec;
        face_patch.head_r.convertTo(rvec, CV_32F);
        face_patch.head_t.convertTo(tvec, CV_32F);
        draw_landmarks(debug_img, rvec, tvec);

        vector<Point3f> cam_points;
        cam_points.push_back(face_patch.face_pos);
        cam_points.push_back(face_patch.face_pos + gaze_length*face_patch.gaze_vec);

        vector<Point2f> img_points;
        projectPoints(cam_points, zero, zero, camera_matrix_, dist_coeffs_, img_points);

        line(debug_img, img_points[0], img_points[1], CV_RGB(255,0,0), 5); // gaze ray
        circle(debug_img, img_points[0], 5, CV_RGB(255,0,0), -1); // staring point
        circle(debug_img, img_points[1], 5, CV_RGB(255,0,0), -1); // end point
    }
}

void HeadPoseEstimator::get_normalized_data(const Mat &img, eyeinfo &eye_info)
{
    Mat HR;
    Rodrigues(eye_info.head_r, HR);

    Mat HT = repeat(eye_info.head_t, 1, 6);

    Mat Fc;
    add(HR*face_model_mat_, HT, Fc);

    Mat le = 0.5*(Fc.col(2) + Fc.col(3)); // left eye
    le.copyTo(eye_info.leye_pos);

    get_normalized_eye(img, HR, le, eye_info.leye_img, eye_info.leye_rot, eye_info.leye_conv);

    Mat re = 0.5*(Fc.col(0) + Fc.col(1)); // right eye
    re.copyTo(eye_info.reye_pos);

    get_normalized_eye(img, HR, re, eye_info.reye_img, eye_info.reye_rot, eye_info.reye_conv);
}

void HeadPoseEstimator::get_normalized_face(const Mat &img, facepatchinfo &facepatch)
{
    Mat HR;
    Rodrigues(facepatch.head_r, HR);

    Mat HT = repeat(facepatch.head_t, 1, 6);

    Mat Fc;
    add(HR*face_model_mat_, HT, Fc);

    Mat face_center = (Fc.col(0) + Fc.col(1) + Fc.col(2) + Fc.col(3) + Fc.col(4) + Fc.col(5)) / 6.0; // face center
    face_center.copyTo(facepatch.face_pos);

    get_normalized_face_patch(img, HR, face_center, facepatch.face_img, facepatch.face_rot, facepatch.face_conv);
}

void HeadPoseEstimator::get_normalized_eye(const Mat &img, const Mat &head_R, const Mat &eye_pos,
                                           Mat &eye_img, Mat &eye_pose, Mat &eye_conv)
{
    int focal_new = 960;
    int distance_new = 600;
    Size roiSize(60,36);

    Mat img_gray;
    cvtColor(img, img_gray, CV_BGR2GRAY);

    float distance = norm(eye_pos);
    float z_scale = distance_new / distance;

    Mat cam_new = (Mat_<float>(3,3) << focal_new, 0, roiSize.width/2, 0, focal_new, roiSize.height/2, 0, 0, 1.0);
    Mat scaleMat = (Mat_<float>(3,3) << 1.0, 0.0, 0.0, 0.0, 1.0, 0.0, 0.0, 0.0, z_scale);

    Mat hRx = head_R.col(0);
    Mat forward = eye_pos/distance;
    Mat down = forward.cross(hRx);
    down = down / norm(down);
    Mat right = down.cross(forward);
    right = right / norm(right);

    Mat rotMat = Mat(3, 3, CV_32F);
    right.copyTo(rotMat.col(0));
    down.copyTo(rotMat.col(1));
    forward.copyTo(rotMat.col(2));
    rotMat = rotMat.t();

    Mat cnvMat = scaleMat * rotMat;
    Mat warpMat = cam_new * cnvMat * camera_matrix_.inv();

    // crop image and copy the equalized image
    warpPerspective(img_gray, eye_img, warpMat, roiSize);
    equalizeHist(eye_img, eye_img);

    // convert and copy the rotation
    Rodrigues(cnvMat*head_R, eye_pose);

    // save the conversion matrix used
    eye_conv = cnvMat;
}

void HeadPoseEstimator::get_normalized_face_patch(const Mat &img, const Mat &head_R, const Mat &face_pos,
                                           Mat &face_img, Mat &face_pose, Mat &face_conv)
{
    int focal_new = 960;
    int distance_new = 300;
    Size roiSize(448,448);

    float distance = norm(face_pos);
    float z_scale = distance_new / distance;

    Mat cam_new = (Mat_<float>(3,3) << focal_new, 0, roiSize.width/2, 0, focal_new, roiSize.height/2, 0, 0, 1.0);
    Mat scaleMat = (Mat_<float>(3,3) << 1.0, 0.0, 0.0, 0.0, 1.0, 0.0, 0.0, 0.0, z_scale);

    Mat hRx = head_R.col(0);
    Mat forward = face_pos/distance;
    Mat down = forward.cross(hRx);
    down = down / norm(down);
    Mat right = down.cross(forward);
    right = right / norm(right);

    Mat rotMat = Mat(3, 3, CV_32F);
    right.copyTo(rotMat.col(0));
    down.copyTo(rotMat.col(1));
    forward.copyTo(rotMat.col(2));
    rotMat = rotMat.t();

    Mat cnvMat = scaleMat * rotMat;
    Mat warpMat = cam_new * cnvMat * camera_matrix_.inv();

    // crop image and copy the equalized image
    warpPerspective(img, face_img, warpMat, roiSize);
    //equalizeHist(face_img, face_img);

    // convert and copy the rotation
    Rodrigues(cnvMat*head_R, face_pose);

    // save the conversion matrix used
    face_conv = cnvMat;
}
