#include <opencv2/core/core.hpp>
#include <opencv2/highgui/highgui.hpp>
#include <opencv2/imgproc/imgproc.hpp>

#include <string>
#include <iostream>

#include "common.h"
#include "facetracker.h"
#include "headposeestimator.h"
#include "gazeestimator.h"
#include "displayspace.h"

#include <boost/program_options.hpp>
#include <boost/filesystem.hpp>
#include <opencv2/ml/ml.hpp>

using namespace std;
using namespace cv;
using namespace cv::ml;
using namespace gaze_data;

unsigned long parse_face_log(const string log_line, faceinfo &face)
{
    stringstream ss(log_line);
    string token;

    vector<double> raw_values;
    while(getline(ss, token, ','))
    {
        raw_values.push_back(stod(token));
    }

    unsigned long frame_id = raw_values[0];
    face.face_id = raw_values[1];
    face.certainty = raw_values[2];

    face.face_bb.x = raw_values[3];
    face.face_bb.y = raw_values[4];
    face.face_bb.width = raw_values[5]-raw_values[3];
    face.face_bb.height = raw_values[6]-raw_values[4];

    for(int p=0; p<6; ++p)
    {
        face.landmarks[p].x = raw_values[7 + 2*p];
        face.landmarks[p].y = raw_values[7 + 2*p + 1];
    }

    return frame_id;
}

int main(int argc, char** argv)
{
    namespace fs = boost::filesystem;
    namespace po = boost::program_options;

    // parse command line options for input/output paths
    po::options_description command_line("Command line options");
    command_line.add_options()
            ("input,i", po::value<string>(), "input path (video/directory) or camera")
            ("output,o", po::value<string>(), "output directory")
            ("calib,c", po::value<string>(), "calibration directory")
            ("config", po::value<string>(), "config file")
            ("debug,d", "show debug output")
            ("reuse,r", "reuse face alignment results saved in the output directory")
            ("facelog,l", po::value<string>(), "facelog dir")
            ("face_model,f", "to use face model or not")
            ("save_video,v", "save output visualization or not")
            ("save_face_feature", "save the extracted face feature or not")
    ;

    cout << "Parsing command line options..." << endl;
    po::variables_map vm_command;
    po::store(po::parse_command_line(argc, argv, command_line), vm_command);
    po::notify(vm_command);

    // parse config file for data paths
    po::options_description config_file("Config file options");
    config_file.add_options()
            ("clm_data_path", po::value<string>(), "CLM data path")
            ("face_model_path", po::value<string>(), "Face model path")
            ("cnn_param_path", po::value<string>(), "Caffe prototxt path")
            ("cnn_model_path", po::value<string>(), "Caffe model path")
            ("input,i", po::value<string>(), "input path (video/directory) or camera")
            ("output,o", po::value<string>(), "output directory")
            ("calib,c", po::value<string>(), "calibration directory")
            ("facelog", po::value<string>(), "facelog dir")
            ("face_model", po::value<bool>(), "face model or not")
            ("save_video", po::value<bool>(), "save output visualization or not")
            ("save_face_feature", po::value<bool>(), "save the extracted face feature or not")
            ("svm_model_path", po::value<string>(), "eye contact detection SVM model path")
    ;

    fs::path exec_root = fs::system_complete(fs::path(argv[0])).parent_path();

    fs::path config_path = exec_root / "../default.cfg";
    if(vm_command.count("config")) config_path = vm_command["config"].as<string>();
    cout << "Reading config from " << config_path.string() << endl;

    if(!fs::exists(config_path))
    {
        cout << "Config file does not exist" << endl;
        exit(EXIT_FAILURE);
    }

    ifstream settings_file(config_path.string());
    po::variables_map vm_config;
    po::store(po::parse_config_file(settings_file , config_file), vm_config);
    po::notify(vm_config);

    if(!vm_config.count("clm_data_path") || !vm_config.count("face_model_path") ||
       !vm_config.count("cnn_param_path") || !vm_config.count("cnn_model_path"))
    {
        cout << "Config file invalid" << endl;
        exit(EXIT_FAILURE);
    }

    string clm_data_path = vm_config["clm_data_path"].as<string>();
    string face_model_path = vm_config["face_model_path"].as<string>();
    string cnn_param_path = vm_config["cnn_param_path"].as<string>();
    string cnn_model_path = vm_config["cnn_model_path"].as<string>();
    bool is_face_model, save_video, save_face_feature;

    bool do_eyecontact_detection = false;
    string svm_model_path;
    if(vm_config.count("svm_model_path"))
    {
        do_eyecontact_detection = true;
        svm_model_path = vm_config["svm_model_path"].as<string>();
    }

    // check input requirements
    fs::path input_path, calib_dir, facelog_dir;

    if(vm_command.count("input"))
        input_path = vm_command["input"].as<string>();
    else if(vm_config.count("input"))
        input_path = vm_config["input"].as<string>();

    // default path is the current working directory
    fs::path output_dir = fs::current_path();
    if(vm_command.count("output"))
        output_dir = vm_command["output"].as<string>();
    else if(vm_config.count("output"))
        output_dir = vm_config["output"].as<string>();

    if(vm_command.count("calib"))
        calib_dir = vm_command["calib"].as<string>();
    else if(vm_config.count("calib"))
        calib_dir = vm_config["calib"].as<string>();

    if(vm_command.count("facelog"))
        facelog_dir = vm_command["facelog"].as<string>();
    else if(vm_config.count("facelog"))
        facelog_dir = vm_config["facelog"].as<string>();
    else
        facelog_dir = output_dir;

    if(vm_command.count("face_model"))
        is_face_model = true;
    else if(vm_config.count("face_model"))
        is_face_model = vm_config["face_model"].as<bool>();
    else
        is_face_model = false;

    if(vm_command.count("save_video"))
        save_video = true;
    else if(vm_config.count("save_video"))
        save_video = vm_config["save_video"].as<bool>();
    else
        save_video = false;

    if(vm_command.count("save_face_feature"))
        save_face_feature = true;
    else if(vm_config.count("save_face_feature"))
        save_face_feature = vm_config["save_face_feature"].as<bool>();
    else
        save_face_feature = false;

    fs::path calib_path = calib_dir / "calibration.yml";
    fs::path monitor_path = calib_dir / "monitor.yml";

    if(!fs::exists(input_path) && !(input_path == "camera"))
    {
        cout << "Input file/directory does not exist" << endl;
        exit(EXIT_FAILURE);
    }

cout << output_dir << endl;
    if(!fs::exists(output_dir)){
        if(fs::create_directories(output_dir)){
            cout << "Create the output directory" << endl;
        }
        else{
            cout << "Cannot create the output directory" << endl;
            exit(EXIT_FAILURE);
        }
    }

    bool save_output = false;
    if(fs::exists(output_dir)) save_output = true;
    else cout << "Output directory does not exist" << endl;

    bool show_debug = false;
    if(vm_command.count("debug") || (input_path == "camera")){
        show_debug = true;
    }

    bool reuse_faces = false;
    if(save_output && vm_command.count("reuse")) reuse_faces = true;

    FaceTracker face_tracker(clm_data_path);
    HeadPoseEstimator pose_estimator(calib_path.string(), face_model_path);
    GazeEstimator gaze_estimator(cnn_param_path, cnn_model_path);
    DisplaySpace display_space(calib_path.string(), monitor_path.string());

    if(show_debug) namedWindow("frame");

    if(fs::is_directory(input_path))
    {
        // assuming image directory
        cout << "process image directory " << input_path << endl;

        int num_frame = 0;
        fs::directory_iterator end_itr;
        for(fs::directory_iterator i(input_path); i!=end_itr; ++i)
        {
            if(!fs::is_regular_file(i->status())) continue;

            fs::path file_path = i->path();

            // only accepts jpg, png or bmp images
            if(file_path.extension() != ".jpg" && file_path.extension() != ".png" && file_path.extension() != ".bmp") continue;

            cout << "process image " << file_path << endl;

            queue<string> face_logs;

            ofstream face_output, gaze_output;
            if(save_output)
            {
                fs::path face_path, gaze_path;
                if (is_face_model)
                {
                    gaze_path = facelog_dir / (file_path.stem().string() + "_face_gaze.csv");
                }
                else
                {
                    gaze_path = facelog_dir / (file_path.stem().string() + "_eye_gaze.csv");
                }

                face_path = facelog_dir / (file_path.stem().string() + "_face_det.csv");

                if(reuse_faces && fs::exists(face_path)) // reuse the detected face results
                {
                    cout << "reuse face alignment results from " << face_path << endl;
                    ifstream face_log(face_path.string());

                    string line;
                    while(getline(face_log, line))
                    {
                        face_logs.push(line);
                    }
                    face_log.close();
                }
                else if(!fs::exists(face_path))
                {
                    cout << "previous face alignment results do not exist: " <<face_path<< endl;
                    reuse_faces = false;
                }

                if(!reuse_faces)
                    face_output.open(face_path.string());
                gaze_output.open(gaze_path.string());
            }

            Mat frame, disp, dis_landmark;
            frame = imread(file_path.string());

            if(show_debug) disp = frame.clone();

            // face tracking using CLM
            vector<faceinfo> faces;
            if(reuse_faces)
            {
                faceinfo face;
                while(!face_logs.empty())
                {
                    faces.push_back(face);
                    face_logs.pop();
                }
            }
            else face_tracker.track_faces(frame, faces);

            // gaze estimation
            vector<gazeinfo> gazes;
            vector<facepatchinfo> facepatches;
            vector<eyeinfo> eyes;
            if (is_face_model){ // for face model
                // 3D pose estimation and eye cropping
                pose_estimator.obtain_facepatches(frame, faces, facepatches);
                // estimate gaze using CNN
                gaze_estimator.estimate(facepatches);
                // project to display coordiantes
                display_space.project_to_display(facepatches, gazes);
                if(show_debug) {
                    // visualize gaze for debug
                    pose_estimator.draw_gaze(disp, facepatches);
                }
            }
            else { // for eye model
                // 3D pose estimation and eye cropping
                pose_estimator.obtain_eyepatches(frame, faces, eyes);
                // estimate gaze using CNN
                gaze_estimator.estimate(eyes);
                // project to display coordiantes
                display_space.project_to_display(eyes, gazes);
                if(show_debug) {
                    // visualize gaze for debug
                    pose_estimator.draw_gaze(disp, eyes);
                }
            }

            ofstream face_feature_file;
            if (save_face_feature){
                fs::path output_path = output_dir / (file_path.stem().string() + "_face_feature.csv");
                cout << "save face feature results to " << output_path << endl;
                face_feature_file.open(output_path.string());
            }


            // write results
            if(save_output)
            {
                Mat frame_copy = frame.clone();
                facepatchinfo facepatch;
                eyeinfo eye;
                for(int i=0; i<gazes.size(); ++i)
                {
                    faceinfo face = faces[i];
                    if (is_face_model)
                        facepatch = facepatches[i];
                    else
                        eye = eyes[i];

                    gazeinfo gaze = gazes[i];

                    // the detected face
                    if(!reuse_faces)
                    {
                        face_output << face.face_id << ",";
                        face_output << face.certainty << ",";
                        face_output << face.face_bb.x << ",";
                        face_output << face.face_bb.y << ",";
                        face_output << face.face_bb.x + face.face_bb.width<< ",";
                        face_output << face.face_bb.y + face.face_bb.height<< ",";
                        for(int p=0; p<6; ++p) // 8-19
                        {
                            face_output << face.landmarks[p].x << ",";
                            face_output << face.landmarks[p].y << ",";
                        }
                        face_output<<endl;
                    }

                    Vec3f head_t, head_r;
                    if (is_face_model){
                        head_t = facepatch.head_t;
                        head_r = facepatch.head_r;
                    }
                    else{
                        head_t = eye.head_t;
                        head_r = eye.head_r;
                    }
                    gaze_output << head_r.val[0] << ","; // head rotation
                    gaze_output << head_r.val[1] << ",";
                    gaze_output << head_r.val[2] << ",";
                    gaze_output << head_t.val[0] << ","; // head translation
                    gaze_output << head_t.val[1] << ",";
                    gaze_output << head_t.val[2] << ",";

                    // here the gaze projection point on the screen coordinate
                    // please refer to the common.h for other outputs you can achieve, e.g. gaze direction in the camera cooridnate.
                    if (is_face_model) {
                        gaze_output << gaze.gaze.x << ",";
                        gaze_output << gaze.gaze.y << endl;
                    }
                    else {
                        gaze_output << gaze.lgaze.x << ",";
                        gaze_output << gaze.lgaze.y << ",";
                        gaze_output << gaze.rgaze.x << ",";
                        gaze_output << gaze.rgaze.y << endl;
                    }

                    if (save_face_feature){
                        for (int num_f=0; num_f<4096; ++num_f){
                        face_feature_file <<facepatch.feature[num_f]<<",";
                        }
                        face_feature_file << endl;
                    }

                }
            }

            // reset tracker status
            face_tracker.reset();

            if(show_debug)
            {
                imshow("frame", disp);

                char key = waitKey(0);
                if (key == 27) break;
            }

            cout << endl;
            if(!reuse_faces && face_output.is_open()) face_output.close();
            if(save_face_feature && face_feature_file.is_open()) face_feature_file.close();
            if(gaze_output.is_open()) gaze_output.close();

        }
    }
    // if the input is a video file
    else if(input_path.extension() == ".avi" || input_path.extension() == ".mp4" || input_path.extension() == ".mkv")
    {
        cout << "process video file " << input_path << endl;

        ofstream face_output, gaze_output;
        queue<string> face_logs;
        if(save_output)
        {
            fs::path face_path, gaze_path;
            if (is_face_model)
            {
                gaze_path = facelog_dir / (input_path.stem().string() + "_face_gaze.csv");
            }
            else
            {
                gaze_path = facelog_dir / (input_path.stem().string() + "_eye_gaze.csv");
            }

            face_path = facelog_dir / (input_path.stem().string() + "_face_det.csv");

            cout << "save results to " << gaze_path << endl;

            if(reuse_faces && fs::exists(face_path))
            {
                cout << "reuse face alignment results from " << face_path << endl;
                ifstream face_log(face_path.string());

                string line;
                while(getline(face_log, line))
                {
                    face_logs.push(line);
                }
                face_log.close();
            }
            else if(!fs::exists(face_path))
            {
                cout << "previous face alignment results do not exist: " <<face_path<< endl;
                reuse_faces = false;
            }

            if(!reuse_faces)
                face_output.open(face_path.string());
            gaze_output.open(gaze_path.string());
        }

        VideoCapture cap(input_path.string());

        if(!cap.isOpened())
        {
            cout << "Could not open video file" << endl;
            exit(EXIT_FAILURE);
        }
        int frame_count = cap.get(CV_CAP_PROP_FRAME_COUNT);
        cout<<"The total number of frames are: "<<frame_count<<endl;
        int frame_count_10_per = int(float(frame_count) / 10.0f);

        unsigned long frame_id = 0;
        int num_frame = 0;

        ofstream face_feature_file;
        if (save_face_feature){
            fs::path output_path = output_dir / (input_path.stem().string() + "_face_feature.csv");
            cout << "save face feature results to " << output_path << endl;
            face_feature_file.open(output_path.string());
        }

        // save the video
        VideoWriter outputVideo;
        if(save_video)
        {
            // Acquire input size
            Size S = Size((int) cap.get(CV_CAP_PROP_FRAME_WIDTH), (int) cap.get(CV_CAP_PROP_FRAME_HEIGHT));
            fs::path out_video_path = output_dir / (input_path.stem().string() + "_det_results.avi");
            cout<<"save video to: "<<out_video_path.string()<<endl;
            int fps = cap.get(CV_CAP_PROP_FPS);
            outputVideo.open(out_video_path.string(), CV_FOURCC('a','v','c','1'), fps, S, true);
            if (!outputVideo.isOpened()) {
                cout << "Could not open the output video file for write\n";
                return -1;
            }
        }

        while (num_frame<=frame_count)
        {
            num_frame++;
            Mat frame, disp;

            if (!cap.read(frame)){
                frame_id++;
                continue; // you never know if it is always success, life is adventure.
            }

            if(save_video) disp = frame.clone();

            // face tracking
            vector<faceinfo> faces;
            if(reuse_faces)
            {
                faceinfo face;
                while(!face_logs.empty() && parse_face_log(face_logs.front(), face) == frame_id)
                {
                    faces.push_back(face);
                    face_logs.pop();
                }
            }
            else face_tracker.track_faces(frame, faces);

            // gaze estimation
            vector<gazeinfo> gazes;
            vector<facepatchinfo> facepatches;
            vector<eyeinfo> eyes;
            if (is_face_model){ // for face model
                // 3D pose estimation and eye cropping
                pose_estimator.obtain_facepatches(frame, faces, facepatches);
                // estimate gaze using CNN
                gaze_estimator.estimate(facepatches);
                // project to display coordiantes
                display_space.project_to_display(facepatches, gazes);
                if(save_video) {
                    // visualize gaze for debug
                    pose_estimator.draw_gaze(disp, facepatches);
                }
            }
            else {
                // 3D pose estimation and eye cropping
                pose_estimator.obtain_eyepatches(frame, faces, eyes);
                // estimate gaze using CNN
                gaze_estimator.estimate(eyes);
                // project to display coordiantes
                display_space.project_to_display(eyes, gazes);
                if(save_video) {
                    // visualize gaze for debug
                    pose_estimator.draw_gaze(disp, eyes);
                }
            }

            // write results
            Mat frame_copy, frame_face_save;
            if(save_output)
            {
                facepatchinfo facepatch;
                eyeinfo eye;
                for(int i=0; i<gazes.size(); ++i)
                {
                    faceinfo face = faces[i];
                    if (is_face_model)
                        facepatch = facepatches[i];
                    else
                        eye = eyes[i];

                    gazeinfo gaze = gazes[i];

                    // face info
                    if(!reuse_faces)
                    {
                        face_output << frame_id << ",";
                        face_output << face.face_id << ",";
                        face_output << face.certainty << ",";
                        face_output << face.face_bb.x << ",";
                        face_output << face.face_bb.y << ",";
                        face_output << face.face_bb.x + face.face_bb.width<< ",";
                        face_output << face.face_bb.y + face.face_bb.height<< ",";
                        for(int p=0; p<6; ++p)
                        {
                            face_output << face.landmarks[p].x << ",";
                            face_output << face.landmarks[p].y << ",";
                        }
                        face_output<<endl;
                    }

                    Vec3f head_t, head_r;
                    if (is_face_model){
                        head_t = facepatch.head_t;
                        head_r = facepatch.head_r;
                    }
                    else{
                        head_t = eye.head_t;
                        head_r = eye.head_r;
                    }
                    gaze_output << head_r.val[0] << ","; // head rotation
                    gaze_output << head_r.val[1] << ",";
                    gaze_output << head_r.val[2] << ",";
                    gaze_output << head_t.val[0] << ","; // head translation
                    gaze_output << head_t.val[1] << ",";
                    gaze_output << head_t.val[2] << ",";

                    // here the gaze projection point on the screen coordinate
                    // please refer to the common.h for other outputs you can achieve, e.g. gaze direction in the camera cooridnate.
                    if (is_face_model) {
                        gaze_output << gaze.gaze.x << ",";
                        gaze_output << gaze.gaze.y << endl;
                    }
                    else {
                        gaze_output << gaze.lgaze.x << ",";
                        gaze_output << gaze.lgaze.y << ",";
                        gaze_output << gaze.rgaze.x << ",";
                        gaze_output << gaze.rgaze.y << endl;
                    }

                    if (save_face_feature){
                        // write the face feature
                        for (int num_f=0; num_f<4096; ++num_f){
                            face_feature_file <<facepatch.feature[num_f]<<",";
                        }
                        face_feature_file << endl;
                    }
                }
            }

            if(save_video) outputVideo << disp;

            if(show_debug){
                imshow("frame", disp);

                char key = waitKey(10);
                if (key == 27) break;
            }

            if(frame_id%frame_count_10_per == 0) cout << frame_id/frame_count_10_per*10 << "% is done."<< endl;

            ++frame_id;
        }

        cout << endl;

        if(cap.isOpened()) cap.release();
        if(outputVideo.isOpened()) outputVideo.release();
        if(!reuse_faces && face_output.is_open()) face_output.close();
        if(save_face_feature && face_feature_file.is_open()) face_feature_file.close();
        if(gaze_output.is_open()) gaze_output.close();
    }
    else if(input_path == "camera"){// This part for the eye contact detection with the camera
        cout << "process live stream from camera" << endl;

        if (!is_face_model){
            cout<<"Camera mode only supports the full-face model" << endl;
            exit(EXIT_FAILURE);
        }

        try{
            // the following are used to fix the camera focal length, you need to install "uvcdynctrl" and "guvcview"
            int setup = system("uvcdynctrl -v -d video1 --set='Focus, Auto' 0"); // fix the focus
            int focal_set = system("uvcdynctrl -s \"Focus (absolute)\" 256"); // set the focus
        }
        catch(int e){ // if we cannot fix the camera focal length
        }

        VideoCapture cap(0); // use the camera
        cap.set(CV_CAP_PROP_FRAME_WIDTH, 1280);
        cap.set(CV_CAP_PROP_FRAME_HEIGHT, 720);

        if(!cap.isOpened())
        {
            cout << "Could not open the camera" << endl;
            exit(EXIT_FAILURE);
        }
        Mat frame, disp;

        // for fps calculation
        bool show_fps = true;
        unsigned int frame_count = 0;
        double fps_tracker = -1.0;
        int64 t0 = 0;

        namedWindow("EyeContactDetection", CV_WINDOW_NORMAL);
        // show the image in full screen mode
        setWindowProperty("EyeContactDetection", CV_WND_PROP_FULLSCREEN, CV_WINDOW_FULLSCREEN);

        Ptr<SVM> svm;
        if(do_eyecontact_detection){
            // load the eye contact detection model
            svm = Algorithm::load<ml::SVM>(svm_model_path);
        }

        while(true){
            frame_count++;

            if (!cap.read(frame)){
                continue; // you never know if it is always success, life is adventure.
            }
            disp = frame.clone();
            //// face tracking
            vector<faceinfo> faces;
            face_tracker.track_faces(frame, faces);
            // gaze estimation
            vector<gazeinfo> gazes;
            vector<facepatchinfo> facepatches;
            vector<eyeinfo> eyes;

            //3D pose estimation and face cropping
            pose_estimator.obtain_facepatches(frame, faces, facepatches);
            //estimate gaze using CNN
            gaze_estimator.estimate(facepatches);

            // Work out the framerate
            if (frame_count % 1 == 0){
                double t1 = cv::getTickCount();
                fps_tracker = 1.0 / (double(t1 - t0) / cv::getTickFrequency());
                t0 = t1;
                frame_count = 0;
            }

            if(do_eyecontact_detection){
                // Eye contact detection
                for(int i=0; i<facepatches.size(); ++i){
                    cv::Mat face_feature = cv::Mat(1, 4096, CV_32F, facepatches[i].feature);
                    float label = svm->predict(face_feature);// get the eye contact detection label
                    if (label==0){ // Eye contact off
                        rectangle(disp, faces[i].face_bb, CV_RGB(255, 0, 0), 5);
                        circle(disp, cv::Point(1100, 100), 100, CV_RGB(255,0,0), -1);
                    }
                    else{ // Eye contact on
                        rectangle(disp, faces[i].face_bb, CV_RGB(0, 255, 0), 5);
                        circle(disp, cv::Point(1100, 100), 100, CV_RGB(0,255,0), -1);
                    }
                }
            }
            else{
                // Visualize gaze directions
                pose_estimator.draw_gaze(disp, facepatches);
            }

            flip(disp, disp, 1); // flip the image to show

            if(show_fps){
                char fpsC[255];
                std::sprintf(fpsC, "%f", fps_tracker);
                string fpsSt("FPS: ");
                fpsSt += fpsC;
                cv::putText(disp, fpsSt, cv::Point(20, 40), CV_FONT_HERSHEY_SIMPLEX, 1, CV_RGB(255, 0, 0));
            }

            imshow("EyeContactDetection", disp);

            char key = waitKey(10);
            if (key == 27) break;
        }
    }

    if(show_debug) destroyAllWindows();
    return 0;
}
